#!/usr/bin/env python3
import contextlib
import functools
import importlib.machinery
import importlib.resources
import importlib.util
import locale
import os
import sqlite3
import sys
from collections.abc import Generator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass
from io import BytesIO, StringIO
from pathlib import Path
from typing import Hashable, Optional, Union

import duckdb
import pandas as pd
import tzlocal
from typedstream.stream import TypedStreamReader

import ica.contact
from ica.contact import ContactRecord, get_contact_records
from ica.exceptions import (
    ContactNotFoundError,
    ContactWithSameNameError,
    ConversationNotFoundError,
    DateRangeInvalidError,
    FormatNotSupportedError,
)

# In order to interpolate the user-specified list of chat identifiers into the
# SQL queries, we must join the list into a string delimited by a common
# symbol, then perform a LIKE comparison on this string within the SQL query;
# this is because named SQL parameters only support string values, rather than
# variable-length sequences
CHAT_IDENTIFIER_DELIMITER = "|"


# The path to the database file for the macOS Messages application
DB_PATH = Path.home() / "Library" / "Messages" / "chat.db"


# The formats supported for file output, represented as a dictionary where each
# key is the format name and the value is the file extension corresponding to
# that format
SUPPORTED_OUTPUT_FORMAT_MAP = {
    "csv": "csv",
    "excel": "xlsx",
    "markdown": "md",
    "json": "json",
}


@dataclass
class DataFrameNamespace:
    """
    The namespace containing the relevant dataframes for the specified user in
    the chat database
    """

    messages: pd.DataFrame
    attachments: pd.DataFrame
    handles: pd.DataFrame


def decode_message_attributedbody(data: bytes) -> str:
    """
    The textual contents of some messages are encoded in a special
    attributedBody column on the message row; this attributedBody value is
    in Apple's proprietary typedstream format, but can be parsed with the
    pytypedstream package (<https://pypi.org/project/pytypedstream/>)
    """
    if data:
        for event in TypedStreamReader.from_data(data):
            # The first bytes object is the one we want; it should be safe to
            # convert back to UTF-8
            if type(event) is bytes:
                return event.decode("utf-8")
    return ""


def get_chat_ids_for_contacts(
    con: sqlite3.Connection, contact_records: Sequence[ContactRecord]
) -> list[str]:
    """
    Find the chat IDs for the chat(s) involving exactly the specified contacts.
    """

    # We need to construct a SQL query that finds chats where:
    # 1. The number of participants matches the number of contacts
    # 2. Each contact is represented by at least one of their handles in the
    #    chat

    contact_checks = []
    query_params = []

    for record in contact_records:
        identifiers = list(record.get_identifiers())
        if not identifiers:
            # If a contact has no identifiers, they can't be in a chat
            return []

        placeholders = ", ".join("?" for _ in identifiers)
        contact_checks.append(
            f"MAX(CASE WHEN handle.id IN ({placeholders}) THEN 1 ELSE 0 END)"
        )
        query_params.extend(identifiers)

    # The query ensures that the total participant count matches the number of
    # contacts, and that the number of *matched* contacts also equals the number
    # of contacts (implying every contact is present)
    query = f"""
    SELECT chat_id
    FROM chat_handle_join
    JOIN handle ON chat_handle_join.handle_id = handle.ROWID
    GROUP BY chat_id
    HAVING COUNT(handle.id) = ?
       AND ({" + ".join(contact_checks)}) = ?
    """

    # Add the count parameters (one for COUNT check, one for SUM check)
    full_params = [len(contact_records)] + query_params + [len(contact_records)]

    return [row[0] for row in con.execute(query, full_params).fetchall()]


def get_messages_dataframe(
    con: sqlite3.Connection,
    chat_ids: Sequence[str],
    contact_records: Sequence[ContactRecord],
    timezone: Optional[str] = None,
) -> pd.DataFrame:
    """
    Return a pandas dataframe representing all messages in a particular
    conversation (identified by the given phone number or email address)
    """
    # If no IANA timezone name is specified, default to the name of the system's
    # local timezone
    if not timezone:
        timezone = tzlocal.get_localzone().key

    chat_ids_placeholder = ", ".join(f"'{cid}'" for cid in chat_ids)

    # Create a mapping from identifier to display name for efficient aggregation
    identifier_to_display_name = {}
    for record in contact_records:
        display_name = ica.contact.get_unique_contact_display_name(
            contact_records, record
        )
        for identifier in record.get_identifiers():
            identifier_to_display_name[identifier] = display_name

    return (
        pd.read_sql_query(
            sql=importlib.resources.files("ica")
            .joinpath(os.path.join("queries", "messages.sql"))
            .read_text()
            .format(chat_ids_placeholder=chat_ids_placeholder),
            con=con,
            parse_dates={"datetime": "ISO8601"},
        )
        # SQL provides each date/time as a Unix timestamp (which is implicitly
        # UTC), but the timestamp is timezone-naive when parsed by pandas; so
        # first, we must add the missing timezone information, then we must
        # convert the datetime to the specified timezone
        .assign(
            datetime=lambda df: df["datetime"]
            .dt.tz_localize("UTC")
            .dt.tz_convert(timezone)
        )
        # Decode any 'attributedBody' values and merge them into the 'text'
        # column
        .assign(
            text=lambda df: df["text"].fillna(
                df.loc[df["text"].isna(), "attributedBody"].apply(
                    decode_message_attributedbody
                )
            )
        )
        # Remove 'attributedBody' column now that it has been merged into the
        # 'text' column
        .drop(columns="attributedBody")
        # Use a regex-based heuristic to determine which messages are reactions
        .assign(
            is_reaction=lambda df: df["text"].str.match(
                r"^(Loved|Liked|Disliked|Laughed at|Emphasized|Questioned|Reacted)"
                r" (“(.*?)”|an \w+|(.*?) to “(.*?)”)$"
            )
        )
        # Convert 'is_from_me' values from integers to proper booleans
        .assign(is_from_me=lambda df: df["is_from_me"].astype(bool))
        # Add sender display name
        .assign(
            sender_display_name=lambda df: df["sender_handle"]
            .map(identifier_to_display_name)
            .fillna("Me")
            .where(df["is_from_me"].eq(False), "Me")
        )
    )


def resolve_sender_identifiers(
    contact_records: Sequence[ContactRecord],
    from_people: Sequence[str],
) -> tuple[bool, set[str]]:
    """
    Resolve the user-supplied 'from_people' filters into a set of allowed
    handle identifiers (e.g. phone numbers, email addresses)
    """
    include_me = False
    allowed_handles: set[str] = set()

    for person_filter in from_people:
        if person_filter.lower() == "me":
            include_me = True
            continue

        if person_filter.lower() == "them":
            for contact in contact_records:
                allowed_handles.update(contact.get_identifiers())
            continue

        # Check against contacts
        matching_contacts = []
        for contact in contact_records:
            # Check name (case-insensitive startswith)
            if contact.first_name.lower().startswith(
                person_filter.lower()
            ) or contact.full_name.lower().startswith(person_filter.lower()):
                matching_contacts.append(contact)
                continue

            # Check identifiers (exact match)
            if any(
                identifier.lower() == person_filter.lower()
                for identifier in contact.get_identifiers()
            ):
                matching_contacts.append(contact)
                continue

        if not matching_contacts:
            raise ContactNotFoundError(f"No contact found matching '{person_filter}'")

        if len({c.id for c in matching_contacts}) > 1:
            raise ContactWithSameNameError(
                f"Multiple contacts found matching '{person_filter}'"
            )

        # Add all identifiers for the matched contact
        allowed_handles.update(matching_contacts[0].get_identifiers())

    return include_me, allowed_handles


def filter_dataframe(
    df: pd.DataFrame,
    contact_records: Sequence[ContactRecord],
    from_date: Optional[str] = None,
    to_date: Optional[str] = None,
    from_people: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    """
    Return a copy of the messages dataframe that has been filtered by the
    user-supplied filters
    """

    # Raise an exception if the 'from' date is after the 'to' date
    if from_date and to_date and pd.Timestamp(from_date) > pd.Timestamp(to_date):
        raise DateRangeInvalidError("Date range is backwards")

    # If no person filter is supplied, or if "all" is specified, return values
    # from all senders
    if not from_people or "all" in (p.lower() for p in from_people):
        return df.pipe(
            lambda df: df[df["datetime"] >= from_date] if from_date else df
        ).pipe(lambda df: df[df["datetime"] <= to_date] if to_date else df)

    include_me, allowed_handles = resolve_sender_identifiers(
        contact_records, from_people
    )

    return (
        df.pipe(
            lambda df: df[
                (df["is_from_me"] & include_me)
                # In the macOS Messages database, the handle_id (and thus
                # sender_handle) for outgoing messages (is_from_me=1) in 1-on-1
                # chats refers to the recipient, not the sender. Therefore, we
                # must explicitly exclude messages from "me" when filtering by a
                # specific contact handle, otherwise we will inadvertently
                # include messages sent TO that contact.
                | (df["sender_handle"].isin(allowed_handles) & ~df["is_from_me"])
            ]
        )
        .pipe(lambda df: df[df["datetime"] >= from_date] if from_date else df)
        .pipe(lambda df: df[df["datetime"] <= to_date] if to_date else df)
    )


def get_attachments_dataframe(
    con: sqlite3.Connection,
    chat_ids: Sequence[str],
    timezone: Optional[str] = None,
) -> pd.DataFrame:
    """
    Return a pandas dataframe representing all attachments in a particular
    conversation (identified by the given phone number)
    """
    chat_ids_placeholder = ", ".join(f"'{cid}'" for cid in chat_ids)

    return (
        pd.read_sql_query(
            sql=importlib.resources.files("ica")
            .joinpath(os.path.join("queries", "attachments.sql"))
            .read_text()
            .format(chat_ids_placeholder=chat_ids_placeholder),
            con=con,
            parse_dates={"datetime": "ISO8601"},
        )
        # Expose the date/time of the message alongside each attachment record,
        # for convenience
        .assign(
            datetime=lambda df: df["datetime"]
            .dt.tz_localize("UTC")
            .dt.tz_convert(timezone)
        )
        .assign(is_from_me=lambda df: df["is_from_me"].astype(bool))
    )


def get_handles_dataframe(
    con: sqlite3.Connection,
    contact_records: Sequence[ContactRecord],
) -> pd.DataFrame:
    """
    Return a pandas dataframe representing the handles associated with the
    participants in a particular conversation. This allows for easy joining with
    the messages dataframe.
    """
    # 1. Collect all identifiers to query
    all_identifiers = set()
    for record in contact_records:
        all_identifiers.update(record.get_identifiers())

    if not all_identifiers:
        return pd.DataFrame(
            columns=pd.Index(
                [
                    "handle_id",
                    "first_name",
                    "last_name",
                    "identifier",
                    "contact_id",
                    "display_name",
                ]
            )
        )

    # 2. Query for handle IDs
    placeholders = ", ".join("?" for _ in all_identifiers)
    query = f"SELECT id, ROWID FROM handle WHERE id IN ({placeholders})"

    # Map identifier string -> handle_id (int)
    # Note: handle.id is the string (phone/email), handle.ROWID is the int
    handle_map = dict(con.execute(query, list(all_identifiers)).fetchall())

    # 3. Build rows for the dataframe
    return pd.DataFrame(
        [
            {
                "handle_id": handle_map[identifier],
                "name": getattr(
                    record,
                    "full_name",
                    f"{record.first_name} {record.last_name}".strip(),
                ),
                "first_name": record.first_name,
                "last_name": record.last_name,
                "identifier": identifier,
                "contact_id": record.id,
                "display_name": ica.contact.get_unique_contact_display_name(
                    contact_records, record
                ),
            }
            for record in contact_records
            for identifier in record.get_identifiers()
            if identifier in handle_map
        ]
    )


def get_dataframes(
    contacts: Sequence[str],
    timezone: Optional[str] = None,
    from_date: Optional[str] = None,
    to_date: Optional[str] = None,
    from_people: Optional[Sequence[str]] = None,
) -> DataFrameNamespace:
    """
    Return all dataframes for a specific macOS Messages conversation
    """
    contact_records = get_contact_records(contacts)

    with sqlite3.connect(f"file:{DB_PATH}?mode=ro", uri=True) as con:
        chat_ids = get_chat_ids_for_contacts(con, contact_records)
        if not chat_ids:
            raise ConversationNotFoundError(
                'No conversation found for the contact(s) "{}"'.format(
                    ", ".join(contacts)
                )
            )

        dfs = DataFrameNamespace(
            messages=get_messages_dataframe(con, chat_ids, contact_records, timezone),
            attachments=get_attachments_dataframe(con, chat_ids, timezone),
            handles=get_handles_dataframe(con, contact_records),
        )
        dfs.messages = filter_dataframe(
            dfs.messages,
            contact_records,
            from_date,
            to_date,
            from_people,
        )
        dfs.attachments = filter_dataframe(
            dfs.attachments,
            contact_records,
            from_date,
            to_date,
            from_people,
        )
        return dfs


def prettify_header_name(
    header_name: Hashable, prettified_label_overrides: Optional[dict[str, str]] = None
) -> Hashable:
    """
    Format the given header name to be more human-readable (e.g. "foo_bar" =>
    "Foo Bar")
    """
    if prettified_label_overrides and header_name in prettified_label_overrides:
        return prettified_label_overrides[header_name]

    if header_name and type(header_name) is str:
        return header_name.replace("_", " ").title()
    else:
        return header_name


def infer_format_from_output_file_path(output: Optional[str]) -> Optional[str]:
    """
    Assuming an explicit format was not provided, infer the format from the
    extension of the given output file path
    """
    if not output:
        return None
    ext = Path(output).suffix[1:]
    if ext not in SUPPORTED_OUTPUT_FORMAT_MAP.values():
        return None
    return ext


def make_dataframe_tz_naive(df: pd.DataFrame) -> pd.DataFrame:
    """
    Convert all of the datetime timstamps in the given dataframe to be
    timezone-naive, including all columns and the index
    """
    return df.pipe(
        lambda df: (
            df.set_index(df.index.tz_localize(None))
            if isinstance(df.index, pd.DatetimeIndex)
            else df
        )
    ).assign(
        **{
            col: df[col].dt.tz_localize(None)
            for col in df.select_dtypes(include=["datetime64[ns, UTC]"])
        }
    )


def prepare_df_for_output(
    df: pd.DataFrame, prettified_label_overrides: Optional[dict[str, str]] = None
) -> pd.DataFrame:
    """
    Prepare the given dataframe for output by prettifying column names,
    stripping timezone details incompatible with Excel, and other normalization
    operations; return the normalized dataframe
    """

    return (
        df.rename(
            # Prettify header column (i.e. textual values in first column)
            index=functools.partial(
                prettify_header_name,
                prettified_label_overrides=prettified_label_overrides,
            ),
            # Prettify header row (i.e. column names)
            columns=functools.partial(
                prettify_header_name,
                prettified_label_overrides=prettified_label_overrides,
            ),
        )
        # Prettify index column name
        .rename_axis(
            index=prettify_header_name(df.index.name, prettified_label_overrides)
        )
        # Make all indices start from 1 instead of 0, but only if the index is
        # the default (rather than a custom column)
        .pipe(lambda df: df.set_index(df.index + 1 if not df.index.name else df.index))
        # Make dataframe timestamps timezone-naive (which is required for
        # exporting to Excel)
        .pipe(lambda df: make_dataframe_tz_naive(df))
    )


def output_results(
    analyzer_df: pd.DataFrame,
    format: Optional[str] = None,
    output: Union[str, StringIO, BytesIO, None] = None,
    prettified_label_overrides: Optional[dict[str, str]] = None,
) -> None:
    """
    Print the dataframe provided by an analyzer module
    """
    # Set the locale to the user's default setting to ensure that numbers are
    # formatted correctly (e.g. with thousands separators); however, if the
    # locale has already been set (e.g. by a test), do not override it
    with contextlib.suppress(locale.Error):
        if locale.getlocale(locale.LC_NUMERIC) == (None, None):
            locale.setlocale(locale.LC_ALL, "")

    is_default_index = not analyzer_df.index.name
    output_df = prepare_df_for_output(
        analyzer_df, prettified_label_overrides=prettified_label_overrides
    )

    if format and format not in {
        *SUPPORTED_OUTPUT_FORMAT_MAP.keys(),
        *SUPPORTED_OUTPUT_FORMAT_MAP.values(),
    }:
        raise FormatNotSupportedError(
            f'The format "{format}" is not supported for output'
        )

    if not format and type(output) is str:
        format = infer_format_from_output_file_path(output)

    output_not_specified: bool = False
    if not output and format in ("excel", "xlsx"):
        output_not_specified = True
        output = BytesIO()
    elif not output:
        output_not_specified = True
        output = StringIO()

    # Keyword arguments passed to any of the to_* output methods
    output_args: dict = {"index": not is_default_index}

    # Output executed DataFrame to correct format
    if format in ("xlsx", "excel"):
        output_df.to_excel(output, **output_args)
    elif format == "csv":
        output_df.to_csv(output, **output_args)
    elif format == "json":
        json_output_df = (
            # Reset index if it's to be included in output
            output_df.reset_index() if output_args.get("index") else output_df
        )
        json_output_df.to_json(output, orient="records", date_format="iso", indent=2)
    elif format in ("md", "markdown"):
        output_df.to_markdown(output, **output_args)
    else:
        (
            output_df
            # When we output the dataframe with to_string(), if the index has a
            # name, it will be displayed on a separate line underneath the line
            # with the column names; this is because space needs to be reserved
            # for the columns axis name; to solve this, we can make the name of
            # the index the name of the columns axis, then remove the name from
            # the index (source: <https://stackoverflow.com/a/43635736/560642>)
            .rename_axis(columns=output_df.index.name)
            .rename_axis(index=None)
            .to_string(
                output,
                index=True,
                line_width=100000,
                # Format numbers with thousands separators
                formatters={
                    col: "{:n}".format
                    for col in output_df.select_dtypes(include="number").columns
                },
            )
        )

    # Print output if no output file path was supplied
    if output_not_specified and type(output) is BytesIO:
        sys.stdout.buffer.write(output.getvalue())
    elif output_not_specified and type(output) is StringIO:
        print(output.getvalue(), flush=True)


@contextmanager
def get_sql_connection(
    dfs: DataFrameNamespace,
) -> Generator[duckdb.DuckDBPyConnection, None, None]:
    """
    Create an in-memory DuckDB database containing all ICA dataframes, and yield
    a connection to that database; using DuckDB over sqlite3 ensures that the
    data is exposed virtually rather than copied, improving performance for
    large conversations
    """
    with duckdb.connect(":memory:") as con:
        con.register("messages", dfs.messages)
        con.register("attachments", dfs.attachments)
        yield con


# Execute an arbitrary SQL query against the database of all ICA dataframes
def execute_sql_query(query: str, con: duckdb.DuckDBPyConnection) -> pd.DataFrame:
    """
    Execute the given arbitrary SQL query, provided a connection to the
    in-memory DuckDB database created by ica.get_sql_connection()
    """
    return con.execute(query).df()



#!/usr/bin/env python3
import ica

# The format to use for all date strings
DATE_FORMAT = "%Y-%m-%d"


def main() -> None:
    """
    Generates a comprehensive breakdown of message totals for every day you and
    the other participants have been messaging in the conversation
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contacts=cli_args.contacts,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_people=cli_args.from_people,
    )

    daily_counts = dfs.messages.assign(
        date=lambda df: df["datetime"].dt.floor("D")
    ).pivot_table(
        index="date",
        columns="sender_display_name",
        values="text",
        aggfunc="count",
        fill_value=0,
    )

    # Ensure all participants (and "Me") are represented as columns
    all_participants: list[str] = sorted(dfs.handles["display_name"].unique())
    expected_cols = ["Me"] + all_participants
    daily_counts = daily_counts.reindex(columns=expected_cols, fill_value=0)

    # Rename columns
    daily_counts = daily_counts.rename(
        columns=lambda col: "#_sent_by_me" if col == "Me" else f"#_sent_by_{col}"
    )

    # Calculate total sent
    daily_counts["#_sent"] = daily_counts.sum(axis=1)

    # Filter days with > 0 messages
    daily_counts = daily_counts[daily_counts["#_sent"] > 0]

    # Reorder columns
    cols = ["#_sent", "#_sent_by_me"] + [
        f"#_sent_by_{name}" for name in all_participants
    ]
    daily_counts = daily_counts[cols]

    ica.output_results(
        daily_counts,
        format=cli_args.format,
        output=cli_args.output,
        prettified_label_overrides={
            f"#_sent_by_{display_name}": f"# Sent By {display_name}"
            for display_name in all_participants
        },
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import datetime

import pandas as pd

import ica


def get_first_message_date(dfs: ica.DataFrameNamespace) -> pd.Timestamp:
    """
    Retrieve the date of the very first message sent in the conversation
    """
    return pd.Timestamp(dfs.messages["datetime"].min().date())


def get_sums_by_day(dfs: ica.DataFrameNamespace) -> pd.DataFrame:
    """
    Calculate the text message sums, grouped by date
    """
    if dfs.messages.empty:
        return pd.DataFrame(
            columns=pd.Index(
                ["message_count", "is_from_me", "is_from_them"], dtype="object"
            )
        ).rename_axis(index="date")
    return (
        dfs.messages.assign(message_count=1)
        .resample("D", on="datetime")
        .sum(numeric_only=True)
        .rename_axis(index="date")
        .assign(is_from_them=lambda df: df["message_count"] - df["is_from_me"])
    )


def get_days_messaged_count(sums_by_day: pd.DataFrame) -> int:
    """
    Retrieve the number of days for which at least one message was sent
    """
    return (sums_by_day["message_count"] > 0).sum()


def get_noreply_count(sums_by_day: pd.DataFrame) -> int:
    """
    Generate a day-by-day breakdown of the number of messages sent where only
    one person sent messages for that day
    """
    is_from_me = sums_by_day["is_from_me"]
    is_from_them = sums_by_day["is_from_them"]
    return len(
        sums_by_day[
            (is_from_me.eq(0) & is_from_them.gt(0))
            | (is_from_me.gt(0) & is_from_them.eq(0))
        ]
    )


def main() -> None:
    """
    Generate a summary of message and reaction counts, by person and in total,
    as well as other insightful metrics
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contacts=cli_args.contacts,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_people=cli_args.from_people,
    )

    first_message_date = get_first_message_date(dfs)
    today = pd.Timestamp(datetime.date.today())
    if cli_args.to_date:
        date_upper_bound = min(
            today, pd.Timestamp(cli_args.to_date).replace(tzinfo=None)
        )
    else:
        date_upper_bound = today

    if pd.isna(first_message_date):
        total_days = 0
    else:
        total_days = (date_upper_bound - first_message_date).days + 1

    sums_by_day = get_sums_by_day(dfs)
    days_messaged_count = get_days_messaged_count(sums_by_day)

    messages_only = dfs.messages[~dfs.messages["is_reaction"]]
    reactions_only = dfs.messages[dfs.messages["is_reaction"]]

    totals_map = {
        "messages": len(messages_only),
        "messages_from_me": messages_only["is_from_me"].sum(),
    }

    # Add per-participant message counts
    message_counts = messages_only["sender_display_name"].value_counts()
    for display_name in dfs.handles["display_name"].unique():
        totals_map[f"messages_from_{display_name.lower()}"] = message_counts.get(
            display_name, 0
        )

    totals_map["reactions"] = len(reactions_only)
    totals_map["reactions_from_me"] = reactions_only["is_from_me"].sum()

    # Add per-participant reaction counts
    reaction_counts = reactions_only["sender_display_name"].value_counts()
    for display_name in dfs.handles["display_name"].unique():
        totals_map[f"reactions_from_{display_name.lower()}"] = reaction_counts.get(
            display_name, 0
        )

    totals_map.update(
        {
            "days_messaged": days_messaged_count,
            "days_missed": total_days - days_messaged_count,
            "days_with_no_reply": get_noreply_count(sums_by_day),
        }
    )

    ica.output_results(
        pd.DataFrame(
            {"metric": tuple(totals_map.keys()), "total": tuple(totals_map.values())},
        ).set_index("metric"),
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import emoji
import pandas as pd

import ica

# List of modifier characters representing the skin tones to be removed from the
# emoji list
skin_tones = {"🏻", "🏼", "🏽", "🏾", "🏿"}


class MostFrequentEmojisCLIArguments(ica.TypedCLIArguments):
    """Additional CLI arguments specific to the most_frequent_emojis analyzer"""

    result_count: int


def main() -> None:
    """
    Generates count data for the top 10 most frequently used emojis across the
    entire conversation
    """
    cli_parser = ica.get_cli_parser()
    cli_parser.add_argument(
        "--result-count",
        type=int,
        default=10,
        help="the number of emoji results to rank",
    )
    cli_args = cli_parser.parse_args(namespace=MostFrequentEmojisCLIArguments())
    dfs = ica.get_dataframes(
        contacts=cli_args.contacts,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_people=cli_args.from_people,
    )

    # Filter out reactions as they are not part of the message text analysis
    messages = dfs.messages[~dfs.messages["is_reaction"]].copy()

    # Create a translation table to remove skin tones
    translation_table = str.maketrans("", "", "".join(skin_tones))

    def extract_emojis(text: str) -> list[str]:
        return [
            item["emoji"].translate(translation_table)
            for item in emoji.emoji_list(text)
        ]

    # Extract emojis from all messages at once
    messages["emoji"] = messages["text"].apply(extract_emojis)

    # Explode the list of emojis so each emoji has its own row
    emoji_data = messages.explode("emoji")

    # Drop rows with no emojis
    emoji_data = emoji_data.dropna(subset=["emoji"])

    # Normalize sender column for pivoting
    # We want columns for "Me" and each participant
    emoji_data["sender_column"] = emoji_data["sender_display_name"]
    emoji_data.loc[emoji_data["is_from_me"], "sender_column"] = "Me"

    # Create a pivot table of counts
    results = pd.crosstab(emoji_data["emoji"], emoji_data["sender_column"])

    # Ensure all participants (and "Me") are represented as columns
    all_participants = sorted(dfs.handles["display_name"].unique())
    expected_cols = ["Me"] + all_participants
    results = results.reindex(columns=expected_cols, fill_value=0)

    # Rename columns to match expected output format
    results = results.rename(
        columns=lambda col: "count_from_me" if col == "Me" else f"count_from_{col}"
    )

    # Calculate total count and insert as first column
    results.insert(0, "count", results.sum(axis=1))

    # Sort by total count and take top N
    # Note: We sort by index (emoji) secondarily to ensure deterministic ordering
    # when counts are tied
    results.index.name = "emoji"
    results = (
        results.sort_values(["count", "emoji"], ascending=[False, True])
        .head(cli_args.result_count)
        .astype(int)
    )

    ica.output_results(
        results,
        format=cli_args.format,
        output=cli_args.output,
        prettified_label_overrides={
            f"count_from_{display_name}": f"Count From {display_name}"
            for display_name in all_participants
        },
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import re

import pandas as pd

import ica


class CountPhrasesArgumentParser(ica.TypedCLIArguments):
    """Additional CLI arguments specific to the count_phrases analyzer"""

    phrases: list[str]
    use_regex: bool
    case_sensitive: bool


def get_phrase_counts(
    messages_df: pd.DataFrame,
    phrases: list[str],
    all_participants: list[str],
    use_regex: bool = False,
    case_sensitive: bool = False,
) -> pd.DataFrame:
    patterns = [phrase if use_regex else re.escape(phrase) for phrase in phrases]
    flags = 0 if case_sensitive else re.IGNORECASE

    # Initialize data dictionary with total counts
    data = {
        "phrase": phrases,
        "count": [
            messages_df["text"].str.count(pattern, flags=flags).sum()
            for pattern in patterns
        ],
        "count_from_me": [
            messages_df[messages_df["is_from_me"]]["text"]
            .str.count(pattern, flags=flags)
            .sum()
            for pattern in patterns
        ],
    }

    # Add counts for each participant
    for participant in all_participants:
        participant_msgs = messages_df[
            messages_df["sender_display_name"] == participant
        ]
        data[f"count_from_{participant}"] = [
            participant_msgs["text"].str.count(pattern, flags=flags).sum()
            for pattern in patterns
        ]

    return pd.DataFrame(data).set_index("phrase")


def main() -> None:
    cli_parser = ica.get_cli_parser()
    cli_parser.add_argument("phrases", nargs="+", help="one or more phrases to count")
    cli_parser.add_argument(
        "--use-regex",
        "-r",
        action="store_true",
        help="if specified, treats phrases as regular expressions",
    )
    cli_parser.add_argument(
        "--case-sensitive",
        "-s",
        action="store_true",
        help="if specified, treats phrases as case-sensitive",
    )
    cli_args = cli_parser.parse_args(namespace=CountPhrasesArgumentParser())
    dfs = ica.get_dataframes(
        contacts=cli_args.contacts,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_people=cli_args.from_people,
    )

    all_participants = sorted(dfs.handles["display_name"].unique())
    results = get_phrase_counts(
        dfs.messages[~dfs.messages["is_reaction"]],
        phrases=cli_args.phrases,
        all_participants=all_participants,
        use_regex=cli_args.use_regex,
        case_sensitive=cli_args.case_sensitive,
    )

    ica.output_results(
        results,
        format=cli_args.format,
        output=cli_args.output,
        prettified_label_overrides={
            **{phrase: phrase for phrase in cli_args.phrases},
            **{
                f"count_from_{display_name}": f"Count From {display_name}"
                for display_name in all_participants
            },
        },
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import glob
import importlib.resources
import os
import sqlite3
from collections import Counter
from collections.abc import Sequence
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import pandas as pd
import phonenumbers

from ica.exceptions import ContactNotFoundError, ContactWithSameNameError

# The glob pattern matching all AddressBook SQL databases to read from
DB_GLOB = (
    Path.home()
    / "Library"
    / "Application Support"
    / "AddressBook"
    / "Sources"
    / "*"
    / "*.abcddb"
)


# The default region for all phone numbers
DEFAULT_PHONE_NUMBER_REGION = "US"


def normalize_phone_number(phone_number: str) -> Optional[str]:
    """
    Normalize the given phone number to the format required for searching in the
    iMessage database
    """
    if not phone_number:
        return None
    return phonenumbers.format_number(
        phonenumbers.parse(phone_number, region=DEFAULT_PHONE_NUMBER_REGION),
        phonenumbers.PhoneNumberFormat.E164,
    )


def normalize_email_address(email_address: str) -> Optional[str]:
    """
    Normalize the given email address to the format required for searching in
    the iMessage database
    """
    if not email_address:
        return None
    normalized_email_address = email_address.strip()
    return normalized_email_address


def normalize_contact_identifier(contact_identifier: str) -> str:
    """
    Normalize the given contact identifier (i.e. full name, phone number, or
    email address) to the format required for searching in the contacts database
    """
    with suppress(phonenumbers.NumberParseException):
        # If the contact string is a valid phone number, we should strip the
        # country code and use the national number (i.e. the phone number minus
        # the country code) for the search; this allows us to find the contact
        # regardless of whether or not the country code is stored in the
        # database (since the contacts database stores each phone number as the
        # user typed it, the country code may or may not be present)
        parsed_phone = phonenumbers.parse(
            contact_identifier, region=DEFAULT_PHONE_NUMBER_REGION
        )
        if phonenumbers.is_valid_number(parsed_phone):
            return str(parsed_phone.national_number)
    return contact_identifier


@dataclass
class ContactRecord:
    # The unique identifier for the contact in the AddressBook database
    id: str
    # The first and last name of the contact
    first_name: str
    last_name: str
    # The phone numbers and email addresses associated with the contact
    phone_numbers: list[str] = field(default_factory=list)
    email_addresses: list[str] = field(default_factory=list)

    @property
    def full_name(self) -> str:
        return f"{self.first_name} {self.last_name}".strip()

    def get_identifiers(self) -> set[str]:
        """
        Retrieve the identifiers for the contact that may be stored on the
        messages database
        """
        return set(self.phone_numbers + self.email_addresses)


def get_contact_records_for_source(
    con: sqlite3.Connection, contact_identifier: str
) -> list[ContactRecord]:
    """
    Fetch the contact records matching the identifier and SQL database
    connection; the user-supplied identifier could be a full name, phone number,
    or email address; multiple contacts may be returned if the identifier is
    ambiguous
    """
    contact_identifier = normalize_contact_identifier(contact_identifier)

    rows = pd.read_sql_query(
        sql=importlib.resources.files("ica")
        .joinpath(os.path.join("queries", "contact.sql"))
        .read_text(),
        con=con,
        params={"contact_identifier": contact_identifier},
    )

    records = []
    if rows.empty:
        return records

    for contact_id, group in rows.groupby("contact_id"):
        first_name = group["ZFIRSTNAME"].iloc[0] or ""
        last_name = group["ZLASTNAME"].iloc[0] or ""

        phone_numbers = (
            group["ZFULLNUMBER"]
            .apply(normalize_phone_number)
            .dropna()
            .unique()
            .tolist()
        )
        email_addresses = (
            group["ZADDRESS"].apply(normalize_email_address).dropna().unique().tolist()
        )

        records.append(
            ContactRecord(
                first_name=first_name,
                last_name=last_name,
                phone_numbers=phone_numbers,
                email_addresses=email_addresses,
                id=str(contact_id),
            )
        )

    return records


def coalesce_contact_records(records: Sequence[ContactRecord]) -> list[ContactRecord]:
    """
    Merge contact records that share at least one identifier (phone number or
    email address)
    """
    # A list of merged contact records
    unique_records: list[ContactRecord] = []
    for record in records:
        found = False
        for unique_record in unique_records:
            # If any identifier overlaps, merge
            if record.get_identifiers() & unique_record.get_identifiers():
                # Merge identifiers
                unique_record.phone_numbers = sorted(
                    set(unique_record.phone_numbers) | set(record.phone_numbers)
                )
                unique_record.email_addresses = sorted(
                    set(unique_record.email_addresses) | set(record.email_addresses)
                )
                # Optionally, update names if needed (e.g., keep the longest, or
                # prefer non-empty)
                if not unique_record.first_name and record.first_name:
                    unique_record.first_name = record.first_name
                if not unique_record.last_name and record.last_name:
                    unique_record.last_name = record.last_name
                unique_record.id = record.id
                found = True
                break
        # If no overlapping record was found, add as new
        if not found:
            unique_records.append(record)
    return unique_records


def get_unique_contact_display_name(
    contact_records: Sequence[ContactRecord], contact_record: ContactRecord
) -> str:
    """
    Determine the unique display name for the given contact record, falling back
    to more specific identifiers if necessary to avoid ambiguity
    """
    # 1. First name
    first_name_count = sum(
        1 for c in contact_records if c.first_name == contact_record.first_name
    )
    if first_name_count == 1:
        return contact_record.first_name

    # 2. Full name
    full_name_count = sum(
        1 for c in contact_records if c.full_name == contact_record.full_name
    )
    if full_name_count == 1:
        return contact_record.full_name

    # 3. Phone number
    if contact_record.phone_numbers:
        phone_number = contact_record.phone_numbers[0]
        phone_number_count = sum(
            1
            for c in contact_records
            if c.phone_numbers and c.phone_numbers[0] == phone_number
        )
        if phone_number_count == 1:
            return phone_number

    # 4. Email address
    if contact_record.email_addresses:
        email_address = contact_record.email_addresses[0]
        email_address_count = sum(
            1
            for c in contact_records
            if c.email_addresses and c.email_addresses[0] == email_address
        )
        if email_address_count == 1:
            return email_address

    return contact_record.full_name or contact_record.first_name or "Unknown"


def validate_contact_records(
    contact_records: Sequence[ContactRecord],
) -> None:
    """
    Validate that at least one contact record was found; if multiple records
    have the same full name, raise an error.
    """
    # Check for duplicate full names
    # Check for duplicate contacts with the same name but different identifiers
    name_counts = Counter(
        (record.first_name, record.last_name) for record in contact_records
    )
    duplicates = [name for name, count in name_counts.items() if count > 1]
    if duplicates:
        # Only report the first duplicate found
        name = duplicates[0]
        raise ContactWithSameNameError(
            f'Multiple contacts found for name "{name[0]} {name[1]}". Please specify a phone number or email address instead.'  # noqa: E501
        )


def get_contact_records(
    contact_identifiers: Sequence[str],
) -> list[ContactRecord]:
    """
    Fetch the attributes for the given contact identifiers; each user-supplied
    identifier could be a full name, phone number, or email address; all
    identifiers may not represent the same contact. Returns a flat list of
    unique ContactRecord objects.
    """
    all_records: list[ContactRecord] = []
    found_identifiers: set[str] = set()

    for db_path in glob.iglob(str(DB_GLOB)):
        with sqlite3.connect(f"file:{db_path}?mode=ro", uri=True) as con:
            for contact_identifier in contact_identifiers:
                records_for_source = get_contact_records_for_source(
                    con, contact_identifier
                )
                if records_for_source:
                    all_records.extend(records_for_source)
                    found_identifiers.add(contact_identifier)

    missing_identifiers = set(contact_identifiers) - found_identifiers
    if missing_identifiers:
        raise ContactNotFoundError(
            "No contact(s) found for: {}".format(", ".join(sorted(missing_identifiers)))
        )

    # Coalesce all records across all identifiers and sources
    unique_records = coalesce_contact_records(all_records)
    validate_contact_records(unique_records)
    return unique_records



#!/usr/bin/env python3

import ica


def main() -> None:
    """
    Generates a full, unedited transcript of every message, including reactions,
    between you and the other participants (attachment files not included)
    """
    cli_args = ica.get_cli_parser().parse_args(namespace=ica.TypedCLIArguments())
    dfs = ica.get_dataframes(
        contacts=cli_args.contacts,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_people=cli_args.from_people,
    )
    ica.output_results(
        dfs.messages.assign(
            timestamp=lambda df: df["datetime"],
            sender=lambda df: df["sender_display_name"],
            is_reaction=lambda df: df["is_reaction"].map({True: "Yes", False: "No"}),
            # U+FFFC is the object replacement character, which appears as the
            # textual message for every attachment
            message=lambda df: df["text"].replace(
                r"\ufffc", "(attachment)", regex=True
            ),
        )
        # Output only the following columns and in this particular order
        .loc[:, ["timestamp", "sender", "is_reaction", "message"]],
        format=cli_args.format,
        output=cli_args.output,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import ica


def main() -> None:
    """
    Execute a SQL query against the database of all ICA dataframes
    """
    parser = ica.get_cli_parser()
    parser.add_argument("query", help="the SQL query to execute")
    cli_args = parser.parse_args()

    dfs = ica.get_dataframes(
        contacts=cli_args.contacts,
        timezone=cli_args.timezone,
        from_date=cli_args.from_date,
        to_date=cli_args.to_date,
        from_people=cli_args.from_people,
    )

    # Execute the query and print the resulting dataframe to stdout
    with ica.get_sql_connection(dfs) as con:
        result_df = ica.execute_sql_query(cli_args.query, con)
        ica.output_results(
            result_df,
            format=cli_args.format,
            output=cli_args.output,
        )


if __name__ == "__main__":
    main()
